#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include "gpio.h"

#define TST_FAIL        1
#define TST_PASS        0

int main() {

	int i, j;
	int err = 0;
	circuit_t *maps;
	int status = TST_PASS;
	int value;
	int count;
	int signal;

	maps = (circuit_t *) malloc(NUM_GPIOS * sizeof(circuit_t));
	if (maps == NULL) {
		printf("Failed to allocate maps!\n");
		err = MEM_ERR;
		goto out;
        } /* if */
	memset(maps, 0, NUM_GPIOS * sizeof(circuit_t));

	/* Init nevers and export pins */
	for (i = 0; i < NUM_GPIOS; i++) {
		if (!pin_is_valid(i)) {
			maps[i].never = i;
			continue;
		} /* if */
		maps[i].avail = gpio_pin_open(i) ? 0 : 1;
		if (!maps[i].avail) {
			printf("Failed to open sysfs entry for pin %d\n", i);
			status = TST_FAIL;
			continue;
		} /* if */

		/* what is active level? */
		maps[i].active_low = gpio_pin_get_active_low(i);
	} /* for */

	/* Main loop: iterate over allowed pins and look for short circuits */
	for (i = 0; i < NUM_GPIOS; i++) {
		if (maps[i].never) continue;
		if (!maps[i].avail) continue;
		
		/* set all valid pins to input, target pin - to output
		 * and zero them
		 */
		for (j = 0; j < NUM_GPIOS; j++) {
			if (maps[j].never) continue;
			if (!maps[j].avail) continue;

			if (i == j) {
				gpio_pin_set_direction(j, DIRECTION_OUT);
			} else {
				gpio_pin_set_direction(j, DIRECTION_IN);
			} /* if */
		} /* for */

		if (maps[i].active_low)
			signal = LEVEL_HIGH;
		else
			signal = LEVEL_LOW;

		/* set signal on target pin */
		gpio_pin_set_level(i, signal);

		/* look for short circuits */
		count = 0;
		for (j = 0; j < NUM_GPIOS; j++) {
			if (maps[j].never) continue;
			if (!maps[j].avail) continue;
			if (j == i) continue;

			value = gpio_pin_get_value(j);
			if (value < 0) {
				printf("Failed to open value file\n");
				status = TST_FAIL;
				break;
			} /* if */

			if (value == signal) {
				(maps[i].pins)[count] = j;
				count++;
			} /* if */
		} /* for */
		if (status == TST_FAIL)
			break;

		maps[i].npins = count;
	} /* for */

	/* close (unexport) gpio pins */
	for (i = 0; i < NUM_GPIOS; i++ ) {
		if (!maps[i].never && maps[i].avail) {
			gpio_pin_close(i);
			maps[i].avail = 0;
		} /* if */
	} /* for */

out:
	if (!err && status == TST_PASS) {
		/* report results: */
	  	printf("====== Circuits map obtained: ======\n");
		for(i = 0; i < NUM_GPIOS; i++) {
			printf("Pin %d: ", i);
			if (maps[i].never) {
			  	printf("not tested as must not be touched\n");
			  	continue;
			} /* if */
			if (maps[i].npins == 0) {
				printf("no short circuits\n");
			} else {
				for (j = 0; j < maps[i].npins; j++) {
					printf("pin %d, ", (maps[i].pins)[j]);
				} /* for */
				printf("\n");
			} /* if */
		} /* for  */
		printf("=========== Test PASSED ===========\n");
		free(maps);
		return TST_PASS;
	} else if (!err && status == TST_FAIL) {
		printf("=========== Test FAILED ===========\n");
		free(maps);
		return TST_FAIL;
	} else {
		printf("=========== Test FAILED ===========\n");
		return TST_FAIL;
	} /* if */
} /* main */
