#!/usr/bin/env python3
# $Id: ddr4_test_window.py 1722 2023-10-30 10:03:58Z bili $

import subprocess
import sys
import time
import json
import os
from multiprocessing import Process
from math import ceil
from math import floor
import datetime
import fcntl


class UnitedFunctions():
	"""parent class for testing memory windows on elbrus processors
	contains functions used for p9, e2c3, e16c"""
	def __init__(self):
		self.proc_name = 'undefined'
		self.tests = ['rl0', 'rr0', 'rl-', 'rr-', 'rl+', 'rr+', 'wl0', 'wr0', 'wl-', 'wr-', 'wl+', 'wr+']
		self.read_regs_with_sicregs_file = True
		self.proc_mc = {}
		self.bytes_tested = [x for x in range(9)]
		self.proc_amount = len(self.get_enabled_proc_array())
		self.cores_amount = self.count_cores()
		self.freq = {}
		self.test_amount = self.proc_amount * self.cores_amount
		self.ranks = {}
		self.ranks_shift = 0
		self.vref = None
		self.test_name = None
		self.mc_settings = {}
		self.info = {}
		self.default_read_delays = {}
		self.default_write_delays = {}
		self.default_write_vref = {}
		self.default_read_vref = {}
		self.read_limits = [0, 0]
		self.write_limits = [0, 0]
		self.tests_to_run = {}
		self.actual_vref_change = None
		self.actual_vref_change_write = None
		self.ecc_regs = {}
		self.opt_start_with_ps = 0
		self.remeasure_zero_flag = False
		self.enable_print_boot_info = True
		self.memtester_timeout = 45
		self.ddr_voltage = 1.2

	def read_sic_reg(self, addr, proc):
		"""reads sic regs using /proc/sicread os elbrus interface"""
		if isinstance(addr, int):
			addr_int = addr
			addr = hex(addr)
		elif addr[-1] == 'w': # addr can be with read format: 0x341w
			addr_int = int(addr[0:-1], 16)
		else:
			addr_int = int(addr, 16)
		while 1:
			cmd = 'echo {} > /proc/sicread'.format(addr)
			subprocess.call(cmd, shell=True)
			if self.read_regs_with_sicregs_file:
				try:
					with open('/proc/sicregs') as file:
						reg_string = file.read()
				except FileNotFoundError:
					self.read_regs_with_sicregs_file = False
					with open('/proc/sicread') as file:
						reg_string = file.read()
			else:
				with open('/proc/sicread') as file:
					reg_string = file.read()
			try:
				reg_string = reg_string.split('\n')[proc].split()
				if reg_string == []:
					return None
			except IndexError: # here if no such proc returning none
				return None
			try: # but here if something wrong -> try again
				if int(reg_string[1]) != proc :
					time.sleep(0.1)
					continue
				if int(reg_string[3][1:-2], 16) != addr_int :
					time.sleep(0.1)
					continue
				res = int(reg_string[4], 16)
				return res
			except IndexError:
				pass

	def write_sic_reg(self, addr, value, proc):
		"""writes sic regs using /proc/sicwrite os elbrus interface"""
		if isinstance(addr, int):
			addr = hex(addr)
		cmd = 'echo {} {} {} > /proc/sicwrite'.format(proc, addr, hex(value))
		subprocess.call(cmd, shell=True)

	def count_cores(self):
		"""returns number of cores in 0 proc based on ST_COREn values"""
		res = 0
		for i in range(16):
			if self.read_sic_reg(0x0100 + 0x4*i, 0) & 1 == 1:
				res += 1
		return res

	def get_enabled_proc_array(self):
		"""tryes to read RT_LN on all processors, if not none - proc present"""
		res = []
		for i in range(4):
			if self.read_sic_reg('0x8', i) is not None:
				res.append(i)
		return res

	def get_enabled_mc_array(self, proc):
		raise NotImplementedError()

	def get_mc_freq(self, mc, proc):
		raise NotImplementedError()

	def read_mc_ecc(self, mc, proc):
		raise NotImplementedError()

	def get_enabled_ranks(self, mc, proc):
		raise NotImplementedError()

	def prepare_mc_to_work(self):
		raise NotImplementedError()

	def get_mc_settings(self, mc, proc):
		raise NotImplementedError()

	def read_all_read_delays(self, mc, proc):
		raise NotImplementedError()

	def read_all_write_delays(self, mc, proc):
		raise NotImplementedError()

	def check_eye_step(self, eye_type, step):
		raise NotImplementedError()

	def count_step_ps(self, step):
		raise NotImplementedError()

	def run_memory_scrubbing(self, test):
		raise NotImplementedError()

	def count_shift_limits(self):
		raise NotImplementedError()

	def read_all_write_vref(self, mc, proc):
		raise NotImplementedError()

	def read_all_read_vref(self, mc, proc):
		raise NotImplementedError()

	def get_default_values(self):
		"""fill internal variables"""
		for proc, mcs in self.proc_mc.items():
			for mc in mcs:
				self.ranks['{}_{}'.format(proc, mc)] = self.get_enabled_ranks(mc, proc)
				self.freq['{}_{}'.format(proc, mc)] = self.get_mc_freq(mc, proc)
				self.default_read_delays['{}_{}'.format(proc, mc)] = self.read_all_read_delays(mc, proc)
				self.default_write_delays['{}_{}'.format(proc, mc)] = self.read_all_write_delays(mc, proc)
				self.default_read_vref['{}_{}'.format(proc, mc)] = self.read_all_read_vref(mc, proc)
				self.default_write_vref['{}_{}'.format(proc, mc)] = self.read_all_write_vref(mc, proc)
				self.mc_settings['{}_{}'.format(proc, mc)] = self.get_mc_settings(mc, proc)
		if self.vref is None:
			freq = max(self.freq.values())
			self.vref = {
				freq >= 3200: 0.055,
				3200 > freq >= 2600: 0.060,
				2600 > freq >= 2400: 0.065,
				2400 > freq >= 2000: 0.068,
				2000 > freq: 		 0.075,
			}[True]

	def run_test(self):
		"""runs all tests from tests_to_run var"""
		for test_name, test_info in self.tests_to_run.items():
			if not self.run_single_test(test_info['test_start'],
							test_info['test_end'],
							test_info['timeout'],
							test_info['test_amount'] if 'test_amount' in test_info else self.test_amount):
				return False
		return True

	def run_single_test(self, test_start_command, test_stop_command, timeout, test_amount):
		"""executes test_amount*test_start_command in shell, timeout seconds later executes test_stop_command"""
		ecc_default = self.read_all_ecc_regs()
		test_success = True
		test_proceses = []
		for _ in range(test_amount):
			test_proceses.append(subprocess.Popen(test_start_command, shell=True, preexec_fn=os.setsid))
		timer = Process(target=time.sleep, args=(timeout, ))
		timer.start()
		try:
			while timer.is_alive() and (not self.check_all_processes_dead(test_proceses)):
				self.check_ecc_errors(ecc_default)
			self.check_ecc_errors(ecc_default)
		except RuntimeError:
			timer.terminate()
			test_success = False
		self.kill_all_processes(test_proceses)
		subprocess.call(test_stop_command, shell=True)
		return test_success

	def read_all_ecc_regs(self):
		ecc_regs = {}
		for proc, mcs in self.proc_mc.items():
			for mc in mcs:
				ecc_regs['{}_{}'.format(proc, mc)] = self.read_mc_ecc(mc, proc)
		return ecc_regs

	def check_ecc_errors(self, ecc_default):
		"""compare current ecc regs with default regs"""
		self.ecc_regs = self.read_all_ecc_regs() # save regs for log
		for proc, mcs in self.proc_mc.items():
			for mc in mcs:
				if self.ecc_regs['{}_{}'.format(proc, mc)] != ecc_default['{}_{}'.format(proc, mc)]:
					print('ECC error detected, processor: {} mc: {}'.format(proc, mc))
					raise RuntimeError("ECC error")

	def check_all_processes_dead(self, test_proceses):
		"""return false if at least one of test_proceses is alive"""
		for test in test_proceses:
			if test.poll() is None:
				return False
		return True

	def kill_all_processes(self, test_proceses):
		for test in test_proceses:
			try:
				os.killpg(os.getpgid(test.pid), 9)
			except ProcessLookupError:
				pass

	def read_results_from_file(self):
		"""safe load memory_eye_test_results.json using backups if nessesary"""
		try:
			file = open('memory_eye_test_results.json', 'r')
		except FileNotFoundError:
			return {}
		try:
			results = json.load(file)
			file.close()
			subprocess.call('cp memory_eye_test_results.json .bkp_memory_eye_test_results.json', shell=True)
			subprocess.call('sync', shell=True)
			return results
		except json.decoder.JSONDecodeError:
			print('results file corrapted, trying to restore from bkp copy...')
		try:
			file = open('.bkp_memory_eye_test_results.json', 'r')
		except FileNotFoundError:
			print('results file corrapted, no bkp file...')
			print('ALL PREV results deleted due to error, starting over...')
			file = open('memory_eye_test_results.json', 'w')
			json.dump({}, file, sort_keys=True, indent=4)
			file.close()
			return {}
		try:
			results = json.load(file)
			file.close()
			print('successfully restored from bkp!')
			subprocess.call('cp .bkp_memory_eye_test_results.json memory_eye_test_results.json', shell=True)
			subprocess.call('sync', shell=True)
			return results
		except json.decoder.JSONDecodeError:
			print('bkp file corrapted...')
			print('ALL PREV results deleted due to error, starting over...')
			file = open('memory_eye_test_results.json', 'w')
			json.dump({}, file, sort_keys=True, indent=4)
			file.close()
			return {}

	def print_help(self):
		print('positional arguments are:')
		print('		<test_name>')
		print('possible arguments are:')
		print('		-mc=<mc_num>     (default - all)')
		print('		-proc=<proc_num>     (default - all)')
		print('		-vref=<vref_change_volts>     (default - depends on freq)')
		print('		-ranks_shift (use it for pcb with memory only in slot 1)')
		print('		-test_amount=<tests thread amount during testing> (default = proc_amount*cores_amount)')
		print('		-byte=<byte number to test> (default = all)')
		print('		-start=<first DQS shift, picoseconds> (default = 0)')
		print('		-re (remeasure first zero window)')
		print('		-no_write_vref (disable vref change in mem module even if supported)')
		print('		-v_ddr=<actual ddr voltage> (default = 1.2)')

	def prepare_arguments(self, args):
		"""argument parser"""
		procs = [i for i in range(self.proc_amount)]
		mc = None
		for argument in args:
			try:
				if argument.startswith('-proc='):
					proc = int(argument[len('-proc='):])
					procs = [proc]
					if proc >= self.proc_amount:
						print('no such proc')
						exit(2)
				elif argument.startswith('-mc='):
					mc = int(argument[len('-mc='):])
				elif argument in ['-h', '-help', '--help', '--h']:
					self.print_help()
					exit(2)
				elif argument.startswith('-vref='):
					self.vref = float(argument[len('-vref='):])
				elif argument.startswith('-test_amount='):
					self.test_amount = int(argument[len('-test_amount='):])
				elif argument.startswith('-byte='):
					self.bytes_tested = [int(argument[len('-byte='):])]
				elif argument.startswith('-ranks_shift'):
					self.ranks_shift = 2
				elif argument.startswith('-test_vref_only'):
					self.tests = ['rl-', 'rr-', 'rl+', 'rr+', 'wl-', 'wr-', 'wl+', 'wr+']
				elif argument.startswith('-start='):
					self.opt_start_with_ps = int(argument[len('-start='):])
				elif argument.startswith('-re'):
					self.remeasure_zero_flag = True
				elif argument.startswith('-no_write_vref'):
					self.remove_vref_on_write()
				elif argument.startswith('-no_print'):
					self.enable_print_boot_info = False
				elif argument.startswith('-v_ddr='):
					self.ddr_voltage = float(argument[len('-v_ddr='):])
				else:
					self.test_name = argument
			except ValueError:
				print('mc, proc and byte must be int, vref must be int/float')
				exit(2)
		if self.test_name is None:
			self.print_help()
			exit(2)
		for proc in procs:
			self.proc_mc[proc] = self.get_enabled_mc_array(proc)
			if mc is not None:
				if mc not in self.proc_mc[proc]:
					print('no such mc')
					exit(2)
				self.proc_mc[proc] = [mc]

	def get_boot_info(self):
		"""stores boot version and all info from bootdata in self.info"""
		try:
			file = open('/proc/bootdata', 'r')
			self.info['boot'] = file.read()
			file.close()
		except FileNotFoundError:
			pass

	def get_linux_version(self):
		"""stores linux version in self.info"""
		self.info['linux'] = subprocess.getoutput('uname -a')

	def get_module_info(self):
		"""trying to get memory module name from i2c, working only with addr 0x50 and 0x51, valid at most pcbs"""
		for i2c_addr in ['0x50', '0x51']:
			# subprocess.call('i2cdump  -y 0 {} > spd_info'.format(i2c_addr), shell=True)
			# module_spec = subprocess.getoutput('decode-dimms -x spd_info')
			# self.info['module_spec_{}'.format(i2c_addr)] = module_spec
			subprocess.call('modprobe i2c-dev', shell=True)
			module_tmp = subprocess.getoutput('i2ctransfer -y 0  w1@0x37 0x0 w1@{} 0x49 r32'.format(i2c_addr))
			subprocess.call('i2ctransfer -y 0  w1@0x36 0x0', shell=True)
			module_tmp = module_tmp.split()
			module_name = ''
			try:
				for i in module_tmp:
					module_name += chr(int(i, 16))
				self.info['module_name_{}'.format(i2c_addr)] = module_name[0:0x15C - 0x149+1]
			except ValueError:
				self.info['module_name_{}'.format(i2c_addr)] = 'unknown (i2c fail addr {})'.format(i2c_addr)
			os_version = subprocess.getoutput('uname -a')
			self.info['os'] = os_version

	def print_info(self):
		"""print test info on screen"""
		for proc, mcs in self.proc_mc.items():
			print(self.proc_name)
			print("$Revision: 1722 $")
			print('test name: {}'.format(self.test_name))
			print("Testing processor: {}; mc: {}; byte: {}".format(
				proc,
				', '.join([str(i) for i in mcs]),
				', '.join([str(i) for i in self.bytes_tested])
			))
			print('mc freq: {}'.format(self.freq))
			print('ranks used: {}'.format(self.ranks))
			print('vref shift: {}'.format(self.vref))
			print('proc amount: {}'.format(self.proc_amount))
			print('cores amount: {}'.format(self.cores_amount))
			for mc in mcs:
				if self.enable_print_boot_info:
					for name, value in self.info.items():
						print('{}: {}'.format(name, value))
				for name, value in self.mc_settings['{}_{}'.format(proc, mc)].items():
					print('{}: {}'.format(name, value))
				for name in self.tests_to_run.keys():
					print('Testing with: {}'.format(name))
				return

	def search_eye(self):
		"""main test function - searches ddr eye in picoseconds"""
		results = self.read_results_from_file()
		if self.test_name not in results:
			results[self.test_name] = {}
		for test in self.tests:
			if (test) in results[self.test_name].keys():
				if results[self.test_name][test] == 0 and self.remeasure_zero_flag:
					pass
				else: continue
			print('testing {} ...'.format(test))
			results[self.test_name]['script_ver'] = "$Id: ddr4_test_window.py 1722 2023-10-30 10:03:58Z bili $"
			results[self.test_name]['proc_name'] = self.proc_name
			results[self.test_name]['freq'] = self.freq
			results[self.test_name]['proc_mc'] = self.proc_mc
			results[self.test_name]['delta_vref'] = self.vref
			results[self.test_name]['ranken'] = self.ranks
			results[self.test_name]['cores'] = self.cores_amount
			results[self.test_name]['procs'] = self.proc_amount
			results[self.test_name]['write_limits'] = self.write_limits
			results[self.test_name]['read_limits'] = self.read_limits
			results[self.test_name]['mc_settings'] = self.mc_settings
			results[self.test_name]['info'] = self.info
			results[self.test_name]['tests_to_run'] = self.tests_to_run
			results[self.test_name]['time'] = datetime.datetime.now().isoformat()
			results[self.test_name]['bytes_tested'] = self.bytes_tested
			results[self.test_name][test] = 0
			file = open('memory_eye_test_results.json', 'w')
			json.dump(results, file, sort_keys=True, indent=4)
			file.close()
			subprocess.call('sync', shell=True)
			step = {'l': -1, 'r': 1}[test[1]]
			limit = {'w': self.write_limits, 'r': self.read_limits}[test[0]]
			opt_st = self.opt_start_with_ps // self.count_step_ps(1)
			# opt_st = {'rl0':5, 'rr0':5,'wl0':5, 'wr0':5,  'rl-':3, 'rr-':3, 'rl+':3, 'rr+':3}[test]
			for i in range(opt_st, 0xff):
				if not (-limit[0] <= i * step <= limit[1]):
					print('dqs shift limit reached: {}'.format(test))
					results[self.test_name][test + '_reason'] = 'dqs_shift_limit'
					break
				if self.check_eye_step(test, i * step):
					results[self.test_name][test] = self.count_step_ps(i)
					print(results[self.test_name][test])
					results[self.test_name][test + 's'] = i
					if self.actual_vref_change is not None:
						results[self.test_name]['actual_vref_change'] = self.actual_vref_change
					if self.actual_vref_change_write is not None:
						results[self.test_name]['actual_vref_change_write'] = self.actual_vref_change_write
					file = open('memory_eye_test_results.json', 'w')
					json.dump(results, file, sort_keys=True, indent=4)
					file.close()
					subprocess.call('sync', shell=True)
				else:
					results[self.test_name][test + '_reason'] = 'ECC'
					results[self.test_name][test + '_ecc_regs'] = self.ecc_regs
					file = open('memory_eye_test_results.json', 'w')
					json.dump(results, file, sort_keys=True, indent=4)
					file.close()
					subprocess.call('sync', shell=True)
					self.run_memory_scrubbing(test)
					break
		print('TEST FINISHED')
		print('       -{}V     vref0    +{}V'.format(self.vref, self.vref))
		print('READ: [{}:{}][{}:{}][{}:{}]'.format(
			int(results[self.test_name]['rl-']), int(results[self.test_name]['rr-']),
			int(results[self.test_name]['rl0']), int(results[self.test_name]['rr0']),
			int(results[self.test_name]['rl+']), int(results[self.test_name]['rr+'])))
		if 'wr+' not in results[self.test_name]:
			print('WRITE: [{}:{}]'.format(
				int(results[self.test_name]['wl0']), int(results[self.test_name]['wr0'])))
		else:
			print('WRITE: [{}:{}][{}:{}][{}:{}]'.format(
				int(results[self.test_name]['wl-']), int(results[self.test_name]['wr-']),
				int(results[self.test_name]['wl0']), int(results[self.test_name]['wr0']),
				int(results[self.test_name]['wl+']), int(results[self.test_name]['wr+'])))


	def run(self, args):
		"""function to call all other funcions"""
		self.prepare_arguments(args)
		print('Gathering info...')
		self.prepare_mc_to_work()
		self.find_tests_location()
		self.get_default_values()
		self.count_shift_limits()
		self.get_module_info()
		self.get_boot_info()
		self.get_linux_version()
		self.check_write_vref_change_available()
		self.print_info()
		return self.search_eye()

	def get_enabled_mc_mask(self, proc):
		"""returns mask of enabled mc in system ie [0,1] -> 0x3"""
		res = 0
		mc_array = self.get_enabled_mc_array(proc)
		for mc in mc_array:
			res+= (1 << mc)
		return res

	def find_tests_location(self):
		"""try to find tests in known places"""
		#stdp paths
		script_location = os.path.dirname(os.path.abspath(__file__))
		par_dir = os.path.abspath(os.path.join(script_location, os.path.pardir))
		par_par_dir = os.path.abspath(os.path.join(par_dir, os.path.pardir))
		memtester_name_stdp = 'mem'
		memtester_path_stdp = par_par_dir + '/mem/test/' + memtester_name_stdp
		#memtester
		if os.path.isfile(memtester_path_stdp):
			test_start = memtester_path_stdp + ' -s 10 -c 9999999 -q > /dev/null'
			test_end = ''
			self.tests_to_run['memtester'] = {'test_start': test_start, 'test_end': test_end, 'timeout': self.memtester_timeout}
		elif os.path.isfile('{}/{}'.format(script_location, 'memtester')):
			test_start = '{}/memtester 10m > /dev/null'.format(script_location)
			test_end = ''
			self.tests_to_run['memtester'] = {'test_start': test_start, 'test_end': test_end, 'timeout': self.memtester_timeout}
		#swim
		swim_name = 'swim_p9' if self.proc_name == 'ELBRUS 8CB' else 'swim'
		if os.path.isfile('{}/{}'.format(script_location, swim_name)):
			test_start = 'printf "10.\n.25E5\n.25E5\n1.E6\n.001\n800\n1\n1334\n1334\n" | {}/{} > /dev/null'.format(script_location, swim_name)
			test_end = ''
			self.tests_to_run['swim'] = {'test_start': test_start, 'test_end': test_end, 'timeout': 45}
		#bebeh stress test
		if self.proc_name == 'ELBRUS 2C3':
			stress_test_name = 'test_link_lnx_e2c3'
			if os.path.isfile('{}/{}'.format(script_location, stress_test_name)):
				test_start = '{}/{} 0 4 4 8 2 3 > /dev/null'.format(script_location, stress_test_name)
				test_end = ''
				self.tests_to_run['stress test by bebeh'] = {'test_start': test_start, 'test_end': test_end, 'timeout': 45}
		#bili stress test
		if self.proc_name in ['ELBRUS 8CB','ELBRUS 2C3', 'ELBRUS 16C']:
			test_file = 'mem_stress_test.ko'
			test_source_files = ['generated_test_p9.S', 'generated_test_e16c.S', 'generated_test_e2c3.S', 'stress_test.c', 'Makefile']
			if not os.path.isfile('{}/{}'.format(script_location, test_file)):
				all_test_sources_exist = True
				for test_source in test_source_files:
					if not os.path.isfile('{}/{}'.format(script_location, test_source)):
						all_test_sources_exist = False
				if all_test_sources_exist:
					print('trying to compile test from source files')
					subprocess.call('make')
			if os.path.isfile('{}/{}'.format(script_location, test_file)):
				mc_tested_masks = []
				for proc in range(self.proc_amount):
					mc_tested_masks.append(self.get_enabled_mc_mask(proc))
				mask_args = ' '.join(['p{}mc={}'.format(i, hex(mc_tested_masks[i])) for i in range(self.proc_amount)])
				# here we set test to test all channels of all processers
				test_start = 'insmod mem_stress_test.ko {} iter=2000 iter_sync=500 size=13 bell=1'.format(mask_args)
				test_end = 'rmmod  mem_stress_test.ko'
				self.tests_to_run['stress_test'] = {'test_start': test_start, 'test_end': test_end, 'timeout': 1}
				self.tests_to_run['stress_test']['test_amount'] = 1
		if self.tests_to_run == {}:
			print('no valid tests found. (memtester/swim/stress_test)')
			exit(2)

	def check_write_vref_change_available(self):
		if not self.proc_supports_write_vref():
			self.remove_vref_on_write()

	def remove_vref_on_write(self):
		for test in ['wl-', 'wr-', 'wl+', 'wr+']:
			try:
				self.tests.remove(test)
			except ValueError:
				pass

	def proc_supports_write_vref(self):
		raise NotImplementedError()


class P9_ddr(UnitedFunctions):
	"""p9-specific functions class"""
	def __init__(self):
		super().__init__()
		self.proc_name = 'ELBRUS 8CB'

	def write_phy_reg(self, addr, value, mc, proc):
		addr = 0x4000 + mc*0x1000 + addr
		self.write_sic_reg(addr, value, proc)

	def read_phy_reg(self, addr, mc, proc):
		addr = 0x4000 + mc*0x1000 + addr
		return self.read_sic_reg(addr, proc)

	def get_enabled_mc_array(self, proc):
		res = []
		for mc in range(4):
			MC_CTL = self.read_sic_reg(0x404 + mc*0x40, proc)
			if MC_CTL & 1:
				res.append(mc)
		return res

	def get_mc_freq(self, mc, proc):
		pwr_mgr_addr = 0x284 if mc < 2 else 0x288
		pwr_mgr = self.read_sic_reg(pwr_mgr_addr, proc)
		clkr = (pwr_mgr >> 2) & 0x3f
		clkf = (pwr_mgr >> 8) & 0x1fff
		clkod = (pwr_mgr >> 21) & 0xf
		return 125 * 4 * (clkf+1) / (clkr+1) / (clkod+1)

	def read_mc_ecc(self, mc, proc):
		addr = 0x400 + mc*0x40
		return self.read_sic_reg(addr, proc)

	def get_enabled_ranks(self, mc, proc):
		DTCR1_addr = 0x204
		RANKEN = self.read_phy_reg(DTCR1_addr, mc, proc) >> 16
		res = []
		for i in range(12):
			if (RANKEN >> i) & 1 == 1:
				res.append(i + self.ranks_shift)
		return res

	def prepare_mc_to_work(self):
		for proc, mcs in self.proc_mc.items():
			for mc in mcs:
				PGCR6_value = self.read_phy_reg(0x28, mc, proc)
				self.write_phy_reg(0x28, PGCR6_value | 1, mc, proc)

	def get_mc_settings(self, mc, proc):
		result = {}
		self.write_phy_reg(0x4DC, self.ranks_shift, mc, proc)
		ZQ1PR = self.read_phy_reg(0x694, mc, proc)
		MC_ODT = self.read_sic_reg(0x040c + 0x40*mc, proc)
		MR1 = self.read_phy_reg(0x184, mc, proc)
		MR2 = self.read_phy_reg(0x188, mc, proc)
		result['ODT_mask_read'] = hex(MC_ODT & 0xffff)
		result['ODT_mask_write'] = hex((MC_ODT >> 16) & 0xffff)
		result['PHY Drive pull-up proc'] = 480 / (((ZQ1PR >> 8) & 0xf) + 1)
		result['PHY Drive pull-down proc'] = 480 / (((ZQ1PR >> 12) & 0xf) + 1)
		result['ODT pull-up impedance proc'] = 480 / (((ZQ1PR >> 16) & 0xf) + 1)
		result['ODT pull-down impedance proc'] = 'Hi-Z'
		result['driver mem'] = {0: 240/7, 1: 240/6}[(MR1 >> 1) & 0x1]
		result['RTT read mem'] = {0: "disabled", 1: 60, 2: 120, 3: 40, 4: 240, 5: 48, 6: 80, 7: 34}[(MR1 >> 8) & 0x7]
		result['RTT write mem'] = {0: result['RTT read mem'], 1: 120, 2: 240, 3: 'Hi-Z'}[(MR2 >> 9) & 0x3]
		return result

	def read_all_read_delays(self, mc, proc):
		DXnLCDLR3_addr = 0x78C
		DXnLCDLR4_addr = 0x790
		DXnLCDLR3_default = self.read_all_lcdl(DXnLCDLR3_addr, mc, proc)
		DXnLCDLR4_default = self.read_all_lcdl(DXnLCDLR4_addr, mc, proc)
		lcdl_current = {}
		for rank in DXnLCDLR3_default.keys():
			lcdl_current[rank] = {**DXnLCDLR3_default[rank], **DXnLCDLR4_default[rank]}
		return lcdl_current

	def read_all_write_delays(self, mc, proc):
		DXnLCDLR1_addr = 0x784
		lcdl_current = self.read_all_lcdl(DXnLCDLR1_addr, mc, proc)
		return lcdl_current

	def move_all_lcdl(self, lcdl_default, step):
		for proc, mcs in self.proc_mc.items():
			for mc in mcs:
				for rank in lcdl_default['{}_{}'.format(proc, mc)].keys():
					RANKIDR_addr = 0x4DC
					self.write_phy_reg(RANKIDR_addr, (rank << 16) | rank, mc, proc)
					for addr, value in lcdl_default['{}_{}'.format(proc, mc)][rank].items():
						self.write_phy_reg(addr, self.move_lcdl(value, step), mc, proc)

	def move_lcdl_of_test_type(self, test, step):
		if test[0] == 'w':
			self.move_all_lcdl(self.default_write_delays, step)
		if test[0] == 'r':
			self.move_all_lcdl(self.default_read_delays, step)
		if test[0] == 'a':
			self.move_all_lcdl(self.default_write_delays, step)
			self.move_all_lcdl(self.default_read_delays, step)

	def move_vref_of_test_type(self, test, vref):
		if test[2] == '0' or test[0] == 'w':
			return
		vref = vref * {'-': -1, '+': 1}[test[2]]
		self.modify_read_vref(vref)

	def check_eye_step(self, test, step):
		self.move_vref_of_test_type(test, self.vref)
		self.move_lcdl_of_test_type(test, step)
		result = self.run_test()
		self.move_lcdl_of_test_type(test, 0)
		self.move_vref_of_test_type(test, 0)
		return result

	def count_step_ps(self, step):
		freq = max(self.freq.values())
		for proc, mcs in self.proc_mc.items():
			for mc in mcs:
				DXnMDLR0_addr = 0x7A0
				mdlr_value_default = self.read_phy_reg(DXnMDLR0_addr, mc, proc)
				break
		return floor(step * 1000000 / freq / ((mdlr_value_default >> 16) & 0x1ff))

	def run_memory_scrubbing(self, test):
		subprocess.call('reboot', shell=True)
		time.sleep(20)

	def count_shift_limits(self):
		for proc_mc in self.default_read_delays.values():
			for rank in proc_mc.values():
				for val in rank.values():
					self.read_limits[0] = max(val & 0x1ff, (val >> 16) & 0x1ff, self.read_limits[0])
					self.read_limits[1] = max(0x1ff-(val & 0x1ff), 0x1ff - ((val >> 16) & 0x1ff), self.read_limits[1])
		for proc_mc in self.default_write_delays.values():
			for rank in proc_mc.values():
				for val in rank.values():
					self.write_limits[0] = max(val & 0x1ff, (val >> 16) & 0x1ff, self.write_limits[0])
					self.write_limits[1] = max(0x1ff-(val & 0x1ff), 0x1ff - ((val >> 16) & 0x1ff), self.write_limits[1])

	def read_all_write_vref(self, mc, proc):
		return

	def read_all_read_vref(self, mc, proc):
		DXnGCR5_addr = 0x714
		DXnGCR8_addr = 0x720
		DXn_step = 0x100
		res = {}
		for addr in [DXnGCR5_addr, DXnGCR8_addr]:
			for byte in self.bytes_tested:
				res[addr] = self.read_phy_reg(addr + DXn_step*byte, mc, proc)
		return res

	def modify_read_vref(self, vref_delta):
		if vref_delta > 0:
			vref_delta_steps = ceil(vref_delta / (self.ddr_voltage*0.00698))
		else:
			vref_delta_steps = floor(vref_delta / (self.ddr_voltage*0.00698))
		if self.actual_vref_change is None:
			self.actual_vref_change = vref_delta_steps*self.ddr_voltage*0.00698
		for proc, mcs in self.proc_mc.items():
			for mc in mcs:
				for addr, value in self.default_read_vref['{}_{}'.format(proc, mc)].items():
					new_value = 0
					for shift in [0, 8, 16, 24]:
						vref = (value >> shift) & 0x3f
						new_vref = vref + vref_delta_steps
						if new_vref < 0:
							new_vref = 0
						new_value |= (new_vref << shift)
					self.write_phy_reg(addr, new_value, mc, proc)

	def move_lcdl(self, lcdl_value, step):
		wdqd = lcdl_value & 0x1ff
		wdqd4 = (lcdl_value >> 16) & 0x1ff
		wdqd += step
		wdqd4 += step
		if wdqd < 0:
			wdqd = 0
		if wdqd > 0x1ff:
			wdqd = 0x1ff
		if wdqd4 < 0:
			wdqd4 = 0
		if wdqd4 > 0x1ff:
			wdqd4 = 0x1ff
		return (wdqd4 << 16) | wdqd

	def read_all_lcdl(self, lcdl_address, mc, proc):
		result = {}
		DXn_step = 0x100
		for rank in self.ranks['{}_{}'.format(proc, mc)]:
			RANKIDR_addr = 0x4DC
			self.write_phy_reg(RANKIDR_addr, (rank << 16) | rank, mc, proc)
			one_rank_res = {}
			for byte in self.bytes_tested:
				addr = lcdl_address + byte * DXn_step
				value = self.read_phy_reg(addr, mc, proc)
				one_rank_res[addr] = value
			result[rank] = one_rank_res
		return result

	def proc_supports_write_vref(self):
		return False

class E16C_ddr(UnitedFunctions):
	"""e16c and e2c3-specific functions class"""
	def __init__(self):
		super().__init__()
		if self.cores_amount == 2:
			self.proc_name = 'ELBRUS 2C3'
		if self.cores_amount == 16:
			self.proc_name = 'ELBRUS 16C'

	def convert_phy_addr(self, addr):
		return addr*2

	def write_phy_reg(self, addr, value, mc, proc):
		addr = self.convert_phy_addr(addr)
		self.write_sic_reg('0x400', mc, proc)  # MC_CH select
		self.write_sic_reg('0x4000', addr, proc)  # DDRREG_CA
		self.write_sic_reg('0x4004w', value, proc)  # DDRREG_D

	def read_phy_reg(self, addr, mc, proc):
		addr = self.convert_phy_addr(addr)
		self.write_sic_reg('0x400', mc, proc)  # MC_CH select
		self.write_sic_reg('0x4000', addr, proc)  # DDRREG_CA
		return self.read_sic_reg('0x4004w', proc)  # DDRREG_D

	def get_enabled_mc_array(self, proc):
		HMU_MIC = self.read_sic_reg('0x0d00', proc)
		res = []
		for i in range(8):
			if HMU_MIC & (1 << (i+24)) != 0:
				res.append(i)
		return res

	def get_mc_freq(self, mc, proc):
		if self.proc_supports_write_vref(): # both bugs fixed in to2
			return self.get_mc_freq_from_registers(mc, proc)
		self.write_sic_reg('0x400', mc, proc)
		self.write_sic_reg('0x450', 0xffff000f, proc)
		self.write_sic_reg('0x450', 0xffff0000, proc)
		time.sleep(2)
		self.write_sic_reg('0x450', 0xc, proc)
		mnt_lo = self.read_sic_reg('0x454', proc)
		mnt_hi = self.read_sic_reg('0x45c', proc)
		freq = round((((mnt_hi & 0xffff)<<32) | mnt_lo) * 2 //1000000, -2)
		return freq

	def get_mc_freq_from_registers(self, mc, proc):
		# e16c one pll for 2 mc, e2c3 one mc - one pll
		if (self.proc_name == 'ELBRUS 16C' and (mc&1) == 1): mc = mc -1
		self.write_sic_reg('0x400', mc, proc) #MC_CH select
		pll1 = self.read_sic_reg('0x0504', proc)
		pll2 = self.read_sic_reg('0x0508', proc)
		pll3 = self.read_sic_reg('0x050c', proc)
		NR = (pll1 & 0xfff) + 1
		OD = ((pll1>>12) & 0x7ff) + 1
		NF = ((pll3 << 32) | pll2) / (2**33)
		Fout = 100 * NF / NR / OD
		return Fout * 4

	def read_mc_ecc(self, mc, proc):
		self.write_sic_reg('0x400', mc, proc)  # MC_CH select
		return self.read_sic_reg('0x440', proc)  # mc_ecc

	def get_enabled_ranks(self, mc, proc):
		res = []
		self.write_sic_reg('0x400', mc, proc)  # MC_CH select
		mc_cfg_value = self.read_sic_reg(0x0418, proc)
		for i in [6, 7, 8, 9]:
			if (mc_cfg_value >> i) & 1 == 1:
				res.append(i-6 + self.ranks_shift)
		return res

	def prepare_mc_to_work(self):
		for proc, mcs in self.proc_mc.items():
			for mc in mcs:
				self.write_sic_reg('0x400', mc, proc)  # MC_CH select
				mc_ctl = self.read_sic_reg('0x404', proc)
				self.write_sic_reg('0x404', (mc_ctl & 0xfffffffd) | (1<<14), proc)  # disable vt-drift, enable write vref change
				self.write_phy_reg(0xd0000, 0, mc, proc)  # enable phy read-write

	def get_mc_settings(self, mc, proc):
		res = {}
		self.write_sic_reg('0x400', mc, proc)
		MC_MOPMa = self.read_sic_reg('0x420', proc)
		MC_MOPMb = self.read_sic_reg('0x424', proc)
		MC_MOPMc = self.read_sic_reg('0x428', proc)
		MC_ODT = self.read_sic_reg('0x40c', proc)
		TxOdtDrvStren = self.read_phy_reg(0x1004D, mc, proc)
		TxEqualizationMode = self.read_phy_reg(0x10048, mc, proc)
		DqDqsRcvCntrl = self.read_phy_reg(0x10043, mc, proc)
		TxImpedanceCtrl0 = self.read_phy_reg(0x10041, mc, proc)
		TxImpedanceCtrl1 = self.read_phy_reg(0x10049, mc, proc)
		TxImpedanceCtrl2 = self.read_phy_reg(0x1004B, mc, proc)
		odt_strenP = TxOdtDrvStren & 0x3f
		odt_strenN = (TxOdtDrvStren >> 6) & 0x3f
		DrvStrenDqP = TxImpedanceCtrl0 & 0x3f
		DrvStrenDqN = (TxImpedanceCtrl0 >> 6) & 0x3f
		DrvStrenEQHiDqP = TxImpedanceCtrl2 & 0x3f
		DrvStrenEQLoDqN = (TxImpedanceCtrl2 >> 6) & 0x3f
		DrvStrenFSDqP = TxImpedanceCtrl1 & 0x3f
		DrvStrenFSDqN = (TxImpedanceCtrl1 >> 6) & 0x3f
		self.dfe_enable = (((DqDqsRcvCntrl >> 2) & 1) == 1)
		rtt = (MC_MOPMa >> 24) & 7
		rtt_wr = (MC_MOPMb >> 9) & 3
		dic = (MC_MOPMa >> 17) & 3
		def count_r(val): return 1/sum([((val >> i) & 1) / rs for i, rs in zip(range(6), [480, 240, 240, 120, 120, 120])])
		def count_rz(val): return 'HiZ' if val == 0 else count_r(val)
		res['TxEqualization (FFE)'] = {0: 'OFF', 1: 'ON-prev value', 2: 'ON-next value'}[TxEqualizationMode]
		res['DFE'] = {False: 'OFF', True: 'ON'}[self.dfe_enable]
		res['DBI_write'] = {0: 'OFF', 1: 'ON'}[(MC_MOPMc >> 27) & 1]
		res['DBI_read'] = {0: 'OFF', 1: 'ON'}[(MC_MOPMc >> 28) & 1]
		res['VTdrift'] = 'OFF'
		res['ODT_mask_read'] = hex(MC_ODT & 0xffff)
		res['ODT_mask_write'] = hex((MC_ODT >> 16) & 0xffff)
		res['PHY Drive pull-up proc EQ_ON weak'] = count_rz(DrvStrenDqP)
		res['PHY Drive pull-down proc EQ_ON weak'] = count_rz(DrvStrenDqN)
		res['PHY Drive pull-up proc EQ_ON weak opposite'] = count_rz(DrvStrenEQHiDqP)
		res['PHY Drive pull-down proc EQ_ON weak opposite'] = count_rz(DrvStrenEQLoDqN)
		res['PHY Drive pull-up proc FULL STR'] = count_rz(DrvStrenFSDqP)
		res['PHY Drive pull-down proc FULL STR'] = count_rz(DrvStrenFSDqN)
		res['ODT pull-up impedance proc'] = count_rz(odt_strenP)
		res['ODT pull-down impedance proc'] = count_rz(odt_strenN)
		res['driver mem'] = {0: 240/7, 1: 240/6}[dic]
		res['RTT read mem'] = {1: 60, 2: 120, 3: 40, 4: 240, 5: 48, 6: 80, 7: 34}[rtt]
		res['RTT write mem'] = {0: res['RTT read mem'], 1: 120, 2: 240, 3: 'Hi-Z'}[rtt_wr]
		return res

	def read_all_read_delays(self, mc, proc):
		res = {}  # res[addr] = value
		for byte in self.bytes_tested:
			for nibble in [0, 1]:
				for rank in self.ranks['{}_{}'.format(proc, mc)]:
					addr = 0x01008C + 0x100*nibble + (byte << 12) + rank
					res[addr] = self.read_phy_reg(addr, mc, proc)
		return res

	def read_all_write_delays(self, mc, proc):
		res = {}  # res[addr] = value
		for byte in self.bytes_tested:
			for nibble in [0, 1]:
				for rank in self.ranks['{}_{}'.format(proc, mc)]:
					addr = 0x0100D0 + 0x100*nibble + (byte << 12) + rank
					res[addr] = self.read_phy_reg(addr, mc, proc)
		return res

	def move_all_lcdl(self, lcdl_default, step):
		for proc, mcs in self.proc_mc.items():
			for mc in mcs:
				for addr, value in lcdl_default['{}_{}'.format(proc, mc)].items():
					self.write_phy_reg(addr, self.move_1_32ui(value, step), mc, proc)

	def move_lcdl_of_test_type(self, test, step):
		if test[0] == 'w':
			self.move_all_lcdl(self.default_write_delays, step)
		if test[0] == 'r':
			self.move_all_lcdl(self.default_read_delays, step)
		if test[0] == 'a':
			self.move_all_lcdl(self.default_write_delays, step)
			self.move_all_lcdl(self.default_read_delays, step)

	def move_vref_of_test_type(self, test, vref):
		if test[2] == '0':
			return
		dac_0_addr = 0x010040
		dac_1_addr = 0x010030 if self.dfe_enable else 0x010040
		vref = vref * {'-': -1, '+': 1}[test[2]]
		dac_addr = {'-': dac_0_addr, '+': dac_1_addr}[test[2]]
		if test[0] == 'w':
			self.modify_write_vref(vref)
		if test[0] == 'r':
			self.modify_read_vref(dac_addr, vref)
		if test[0] == 'a':
			self.modify_write_vref(vref)
			self.modify_read_vref(dac_addr, vref)

	def check_eye_step(self, test, step):
		self.move_vref_of_test_type(test, self.vref)
		self.move_lcdl_of_test_type(test, step)
		result = self.run_test()
		self.move_lcdl_of_test_type(test, 0)
		self.move_vref_of_test_type(test, 0)
		return result

	def count_step_ps(self, step):
		freq = max(self.freq.values())
		return floor(step * 1000000 / freq / 32)

	def run_memory_scrubbing(self, test):
		if test[0] == 'w':
			subprocess.call('reboot', shell=True)
			time.sleep(20)
		for proc, mcs in self.proc_mc.items():
			for mc in mcs:
				self.write_sic_reg('0x400', mc, proc)  # MC_CH select
				mcerrrec = self.read_sic_reg('0x438', proc)
				mcerrrec_new = mcerrrec | (1 << 15) | (0xf << 16)
				self.write_sic_reg('0x438', mcerrrec_new, proc)
				time.sleep(30)
				self.write_sic_reg('0x438', mcerrrec, proc)

	def count_shift_limits(self):
		for dd in self.default_read_delays.values():
			for val in dd.values():
				self.read_limits[0] = max(val & 0x1f, self.read_limits[0])
				self.read_limits[1] = max(0x1f - (val & 0x1f), self.read_limits[1])
		for dd in self.default_write_delays.values():
			for val in dd.values():
				self.write_limits[0] = max(val & 0x1f, self.write_limits[0])
				self.write_limits[1] = max(0x1f - (val & 0x1f), self.write_limits[1])

	def move_1_32ui(self, value, step):
		coarse = (value >> 6) & 0xf
		fine = value & 0x1f
		fine += step
		if fine >= 0x20:
			fine = 0x1f
		if fine < 0:
			fine = 0
		return (coarse << 6) | fine

	def read_all_write_vref(self, mc, proc):
		phytid = self.read_phy_reg(0x02001E, mc, proc)
		if phytid == 0:
			self.remove_vref_on_write()
			return None
		return phytid

	def read_all_read_vref(self, mc, proc):
		res = {}
		for dac_addr in [0x010040, 0x010030]:
			res_dac_only = {}
			for byte in self.bytes_tested:
				for bit in range(8):
					addr = dac_addr + 0x100*bit + (byte << 12)
					res_dac_only[addr] = self.read_phy_reg(addr, mc, proc)
			res[dac_addr] = res_dac_only
		DqDqsRcvCntrl = self.read_phy_reg(0x010043, mc, proc)
		MajorMode = (DqDqsRcvCntrl >> 4) & 7
		ExtVrefRange = (DqDqsRcvCntrl >> 1) & 1
		dac_mode = ((MajorMode << 1) | ExtVrefRange)
		res['dac_mode'] = dac_mode
		return res

	def count_steps_in_delta(self, vref_delta, value_step):
		if vref_delta > 0:
			vref_delta_steps = ceil(vref_delta / value_step)
		else:
			vref_delta_steps = floor(vref_delta / value_step)
		return vref_delta_steps

	def modify_read_vref(self, dac_addr, vref_delta):
		VDDQ = self.ddr_voltage
		for proc, mcs in self.proc_mc.items():
			for mc in mcs:
				default_vref = self.default_read_vref['{}_{}'.format(proc, mc)][dac_addr]
				dac_mode = self.default_read_vref['{}_{}'.format(proc, mc)]['dac_mode']
				value_step = {
									6: VDDQ*0.00345,
									7: VDDQ*0.00385
								}[dac_mode]
				vref_delta_steps = self.count_steps_in_delta(vref_delta, value_step)
				if self.actual_vref_change is None:
					self.actual_vref_change = vref_delta_steps * value_step
				for addr, value in default_vref.items():
					self.write_phy_reg(addr, value + vref_delta_steps, mc, proc)

	def modify_write_vref(self, vref_delta):
		VDDQ = self.ddr_voltage
		for proc, mcs in self.proc_mc.items():
			for mc in mcs:
				default_vref = self.default_write_vref['{}_{}'.format(proc, mc)]
				self.write_sic_reg('0x400', mc, proc)  # MC_CH select
				MC_MOPMd = self.read_sic_reg('0x042c', proc)
				MC_MOPMd_value = (MC_MOPMd & 0xffffff00) | (1<<7)
				vref_max = (default_vref >> 8) & 0xff
				vref_min = default_vref & 0xff
				vref_mid = int((vref_max + vref_min)/2)
				vref = {
					(vref_delta > 0): vref_min,
					(vref_delta < 0): vref_max,
					(vref_delta == 0): vref_mid,
				}[True]
				value_step = VDDQ*0.0065
				vref_delta_steps = self.count_steps_in_delta(vref_delta, value_step)
				if self.actual_vref_change_write is None:
					self.actual_vref_change_write = f'{abs(vref_delta_steps * value_step)}/{(abs(vref_delta_steps)+vref_max-vref_min) * value_step}'
				self.write_sic_reg('0x042c', MC_MOPMd_value | (vref + vref_delta_steps), proc)

	def proc_supports_write_vref(self):
		with open('/proc/cpuinfo', 'r') as cpu_info:
			for line in cpu_info:
				if line.startswith('revision'):
					revision = int(line.split()[2])
					if revision >= 1:
						return True
		return False

##########################################
# all above is copy of ddr_search_window #
##########################################

class TestWindowFunctions(UnitedFunctions):
	def __init__(self):
		super().__init__()
		self.tests = ['rl0', 'rr0', 'wl0', 'wr0']
		self.test_name = 'ddr4 test window'

	def search_eye(self):
		freq = max(self.freq.values())
		eye = { freq >= 3200: 36,
				3200 > freq >= 2900: 39,
				2900 > freq >= 2600: 43,
				2600 > freq >= 2400: 42,
				2400 > freq >= 2000: 50,
				2000 > freq        : 60,
				}[True]
		test_status = 'PASSED'
		for test in self.tests:
			print('testing {} ... '.format(test), end='')
			step = {'l':-1, 'r': 1}[test[1]]
			limit = {'w':self.write_limits, 'r': self.read_limits, 'a': self.read_limits}[test[0]]
			i = ceil(eye / self.count_step_ps(1)) + 1
			if not (-limit[0] <= i * step <= limit[1]):
				print('dqs shift limit reached: {}'.format(test))
			if self.check_eye_step(test, i * step):
				print('OK')
			else:
				test_status = 'FAILED'
				break
		print('TEST {}'.format(test_status))
		ps_step = self.count_step_ps(1)
		print('Measurment accuracy: +/-{}'.format(ps_step))
		if self.actual_vref_change != None:
			print('Actual performed vref change: +/{}'.format(self.actual_vref_change))
		print('Actual performed DQS shift:')
		print('READ: [{}:{}]'.format(i*ps_step,i*ps_step))
		print('WRITE: [{}:{}]'.format(i*ps_step,i*ps_step))
		print('Windows at least:')
		print('READ: [{}:{}]'.format(eye,eye))
		print('WRITE: [{}:{}]'.format(eye,eye))
		return(0 if test_status == 'PASSED' else 1)

	def print_help(self):
		print('possible arguments are:')
		print('		-mc=<mc_num>     (default - all)')
		print('		-proc=<proc_num>     (default - all)')
		print('		-vref=<vref_change_volts>     (default - depends on freq)')
		print('		-ranks_shift (use it for pcb with memory only in slot 1)')
		print('		-test_amount=<tests thread amount during testing> (default = proc_amount*cores_amount)')
		print('		-no_print (disable printing info about boot, linux, settings, etc)')
		print('		-fast (speed up test, but less accurate diagnostics if error)')
		print('		-byte=<byte number to test> (default = all)')

	def prepare_arguments(self, args):
		super().prepare_arguments(args)
		for argument in args:
			if argument.startswith('-fast'):
				self.tests = ['al0', 'ar0']
				self.memtester_timeout = 10

class P9_ddr_test(TestWindowFunctions, P9_ddr):
	def __init__(self):
		super().__init__()

class E16C_ddr_test(TestWindowFunctions, E16C_ddr):
	def __init__(self):
		super().__init__()



if __name__ == "__main__":
	fp = open(os.path.realpath(__file__), 'r')
	try:
		fcntl.flock(fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
	except IOError:
		print('multiple launches of script detected')
		exit(2)
	args = sys.argv[1:]
	try:
		test = {
			2: E16C_ddr_test(),
			8: P9_ddr_test(),
			16: E16C_ddr_test(),
		}[UnitedFunctions().cores_amount]
	except KeyError:
		print("Error: not all cores enabled. All cores must be enabled for testing.")
		exit(1)
	exit(test.run(args)) # returns 0 if all ok, else 1
